use bevy::{prelude::*, window::PrimaryWindow};
use bevy_inspector_egui::quick::WorldInspectorPlugin;

const SNAKE_HEAD_COLOR: Color = Color::srgb(0.7, 1.0, 0.7);
const ARENA_WIDTH: u32 = 20;
const ARENA_HEIGHT: u32 = 15;
const WINDOW_WIDTH: f32 = 1280.0;

#[derive(Eq, PartialEq, Copy, Clone)]
enum Direction {
    Up,
    Down,
    Left,
    Right,
}

impl Direction {
    fn opposite(self) -> Self {
        match self {
            Self::Left => Self::Right,
            Self::Right => Self::Left,
            Self::Up => Self::Down,
            Self::Down => Self::Up,
        }
    }
}

#[derive(Component)]
struct SnakeHead {
    direction: Direction,
}

#[derive(Component)]
struct Position {
    x: i32,
    y: i32,
}

#[derive(Component)]
struct Size {
    width: f32,
    height: f32,
}

impl Size {
    pub fn square(x: f32) -> Self {
        Self {
            width: x,
            height: x,
        }
    }
}

#[derive(Resource)]
struct MovementTimer(Timer);

fn snake_movement_input(input: Res<ButtonInput<KeyCode>>, mut heads: Query<(&mut SnakeHead)>) {
    if let Some(mut head) = heads.iter_mut().next() {
        let dir: Direction = if input.pressed(KeyCode::ArrowUp) {
            Direction::Up
        } else if input.pressed(KeyCode::ArrowDown) {
            Direction::Down
        } else if input.pressed(KeyCode::ArrowLeft) {
            Direction::Left
        } else if input.pressed(KeyCode::ArrowRight) {
            Direction::Right
        } else {
            head.direction
        };
        // allow all inputs except turning around
        if dir != head.direction.opposite() {
            head.direction = dir;
        }
    }
}

fn snake_movement(
    input: Res<ButtonInput<KeyCode>>,
    time: Res<Time>,
    mut timer: ResMut<MovementTimer>,
    mut heads: Query<(&SnakeHead, &mut Position)>,
) {
    if let Some((head, mut pos)) = heads.iter_mut().next() {
        if timer.0.tick(time.delta()).just_finished() {
            if head.direction == Direction::Up {
                pos.y += 1;
            } else if head.direction == Direction::Down {
                pos.y -= 1;
            } else if head.direction == Direction::Left {
                pos.x -= 1;
            } else if head.direction == Direction::Right {
                pos.x += 1;
            }
        }
    }
}

fn scale_translation(
    mut windows: Query<&mut Window, With<PrimaryWindow>>,
    mut q: Query<(&Size, &mut Transform)>,
) {
    let window = windows.single_mut();
    for (sprite_size, mut transform) in q.iter_mut() {
        transform.scale = Vec3::new(
            sprite_size.width / ARENA_WIDTH as f32 * window.width(),
            sprite_size.height / ARENA_HEIGHT as f32 * window.height(),
            1.0,
        );
    }
}

fn position_translation(
    mut windows: Query<&mut Window, With<PrimaryWindow>>,
    mut q: Query<(&Position, &mut Transform)>,
) {
    // We subtract half the window width because our coordinate system starts at the bottom left, and
    // Translation starts from the center. We then add half the size of a single tile, because we want
    // our sprites bottom left corner to be at the bottom left of a tile, not the center.
    fn convert(pos: f32, bound_window: f32, bound_game: f32) -> f32 {
        let tile_size = bound_window / bound_game;
        pos / bound_game * bound_window - (bound_window / 2.) + (tile_size / 2.)
    }

    let window = windows.single_mut();
    for (pos, mut transform) in q.iter_mut() {
        transform.translation = Vec3::new(
            convert(pos.x as f32, window.width(), ARENA_WIDTH as f32),
            convert(pos.y as f32, window.height(), ARENA_HEIGHT as f32),
            0.0,
        );
    }
}

fn setup_player(mut commands: Commands) {
    commands
        .spawn(Sprite {
            color: SNAKE_HEAD_COLOR,
            custom_size: Some(Vec2::new(1.0, 1.0)),
            ..default()
        })
        .insert(SnakeHead {
            direction: Direction::Right,
        })
        .insert(Position { x: 0, y: 0 })
        .insert(Size::square(1.0));
}

fn setup_camera(mut commands: Commands) {
    commands.spawn((Camera2d {},));
}

pub struct SnakePlugin;
impl Plugin for SnakePlugin {
    fn build(&self, app: &mut App) {
        app.add_systems(Startup, (setup_camera, setup_player));
        app.add_systems(Update, snake_movement_input);
        app.add_systems(FixedUpdate, snake_movement);
        app.add_systems(PostUpdate, (scale_translation, position_translation));

        app.insert_resource(MovementTimer(Timer::from_seconds(
            1.0,
            TimerMode::Repeating,
        )));
    }
}

fn main() {
    App::new()
        .add_plugins(
            DefaultPlugins.set(WindowPlugin {
                primary_window: Some(Window {
                    title: "Snake".to_string(),
                    resolution: (
                        WINDOW_WIDTH,
                        WINDOW_WIDTH * (ARENA_HEIGHT as f32 / ARENA_WIDTH as f32),
                    )
                        .into(),
                    ..default()
                }),
                ..default()
            }),
        )
        .add_plugins(WorldInspectorPlugin::new())
        .add_plugins(SnakePlugin)
        .run();
}
